/*
   USB connection library
   by paul69, v0.15, 2004-03-27, <e2500@narod.ru>

   Supported camera:
      E995 (tested)
      E2500 (tested)
      E4300 (tested)
      E4500 (tested)
*/
#define STRICT
#include <windows.h>
#pragma hdrstop

#include <stdio.h>
#include "dsc_connect.h"

#ifndef IOCTL_SCSI_PASS_THROUGH_DIRECT
#define IOCTL_SCSI_PASS_THROUGH_DIRECT 0x0004D014
#endif

#ifndef IOCTL_SCSI_PASS_THROUGH
#define IOCTL_SCSI_PASS_THROUGH 0x0004D004
#endif

VENDOR_ENUM default_vendor = NIKON;

enum DSC_COMMAND_ENUM
{
   DSC_SET_REG,   // 0
   DSC_GET_REG,   // 1
   DSC_COMMAND,   // 2
   DSC_SET_VAR,   // 3
   DSC_GET_VAR,   // 4
};

enum USB_COMMAND_ENUM
{
   READ_DEVICE_NAME, //12
   USB_START_IO,     //E0
   USB_SEND_CMD,     //E1
   USB_RCV_RESPONCE, //E2
   USB_RCV_ACK,      //E3
   USB_RCV_SIZE,     //E4
   USB_CLOSE,        //E5
};

struct TScsiCommand
{
   byte command;
   byte data_in;
   word timeout;
};

static TScsiCommand inquiry_command = { 0x12, 1, 12 };

static byte NIKON_dsc_command[] = { 0, 1, 2, 3, 4 };
static TScsiCommand NIKON_command_table[] =
{
   { 0x12, 1, 12 },
   { 0xE0, 0, 2 },
   { 0xE1, 0, 2 },
   { 0xE2, 1, 50 },
   { 0xE3, 1, 2 },
   { 0xE4, 1, 50 },
   { 0xE5, 0, 2 },
};

static byte OLYMPUS_dsc_command[] = { 0, 1, 2, 3, 6 };
static TScsiCommand OLYMPUS_command_table[] =
{
   { 0x12, 1, 5 },
   { 0xC0, 0, 5 },
   { 0xC1, 0, 5 },
   { 0xC2, 1, 5 },
   { 0xC3, 1, 5 },
   { 0xC4, 1, 5 },
   { 0xC5, 0, 5 },
};


int timeout;

class USB_DSC : public DSC
{
public:

   USB_DSC(const char* _comment, VENDOR_ENUM vendor) :
      DSC(_comment)
   {
      if( vendor == UNKNOWN_VENDOR )
         vendor = default_vendor;

      if( vendor == NIKON )
      {
         command_table = NIKON_command_table;
         dsc_command = NIKON_dsc_command;
      }

      if( vendor == OLYMPUS )
      {
         command_table = OLYMPUS_command_table;
         dsc_command = OLYMPUS_dsc_command;
      }
   }

   virtual bool SetReg(int reg, unsigned value) const;
   virtual bool GetReg(int reg, unsigned& value) const;

   virtual bool SetVar(int reg, const void* data, unsigned size) const;
   virtual bool SetStr(int reg, const char*) const;
   virtual bool GetVar(int reg, TResponce&) const;

   virtual bool SpecCmd(int reg, unsigned value, unsigned timeout) const;

   virtual bool GetDeviceInfo(TDeviceInfo&) const;

protected:
   bool CloseCamera() const;

private:
   virtual bool SendCmd(USB_COMMAND_ENUM i_command, void* buf, unsigned size) const =0;

   bool usbStartIO() const;
   bool usbSendCmd(DSC_COMMAND_ENUM type, int reg) const;
   bool usbSendCmdSetStr(int reg, const char* str) const;
   bool usbSendCmdWrite(int reg, const void* data, unsigned sz) const;
   bool usbSendCmdWrite2(int cmd, const void* data, unsigned sz) const;
   bool usbSendCmdSetReg(int reg, unsigned value) const;
   bool usbSpecCmd(int reg, int value) const;
   bool usbSendAck() const;
   bool usbAcceptResponce(void* buf, int size) const;
   bool usbGetAck() const;
   bool usbGetSize(unsigned& size) const;


protected:
   TScsiCommand* command_table;
   byte* dsc_command;
};


////////////////////////////////////////////////////////////////////////////////
//
//
//
//
//

struct TCdbCmd
{
   dword command;
   dword size;
};

struct TCdbCmdUSB
{
   dword command;
   dword _zero;

   byte size_hi;
   byte size_lo;
};

struct TCmdAck
{
   dword length;
   word code; // 1,2,3
   word code2; // 0,0x9FFF
   dword _unknown[14]; // for sizeof == 64
   byte ack;
};

struct TInquiryData
{
    byte DeviceType : 5;
    byte DeviceTypeQualifier : 3;
    byte DeviceTypeModifier : 7;
    byte RemovableMedia : 1;
    byte Versions;
    byte ResponseDataFormat : 4;
    byte HiSupport : 1;
    byte NormACA : 1;
    byte ReservedBit : 1;
    byte AERC : 1;
    byte AdditionalLength;
    byte Reserved[2];
    byte SoftReset : 1;
    byte CommandQueue : 1;
    byte Reserved2 : 1;
    byte LinkedCommands : 1;
    byte Synchronous : 1;
    byte Wide16Bit : 1;
    byte Wide32Bit : 1;
    byte RelativeAddressing : 1;
    byte VendorId[8];
    byte ProductId[16];
    byte ProductRevisionLevel[4];
};

bool IsNikonDSC(TInquiryData& inquiry)
{
   char temp[20];

   memcpy(temp, inquiry.VendorId, 8);
   temp[8] = '\0';

   if( stricmp(temp, "NIKON   ") != 0 ) return false;

   memcpy(temp, inquiry.ProductId, 16);
   temp[16] = '\0';

   if( strstr(temp, "DSC") != NULL ) return true;
   if( strstr(temp, "Digital Camera") != NULL ) return true;

   return false;
}

bool IsOlympusDSC(TInquiryData& inquiry)
{
   char temp[20];

   memcpy(temp, inquiry.VendorId, 8);
   temp[7] = '\0';

   return 0 == stricmp(temp, "OLYMPUS");
}

bool USB_DSC::GetDeviceInfo(TDeviceInfo& info) const
{
   memset( &info, 0, sizeof(info) );

   TInquiryData inquiry;

   if( !SendCmd(READ_DEVICE_NAME, &inquiry, sizeof(inquiry)) )
      return false;

   memcpy( info.VendorId, inquiry.VendorId, sizeof(inquiry.VendorId) );
   info.VendorId[sizeof(inquiry.VendorId)] = '\0';

   memcpy( info.ProductId, inquiry.ProductId, sizeof(inquiry.ProductId) );
   info.ProductId[sizeof(inquiry.ProductId)] = '\0';

   memcpy( info.ProductRevisionLevel, inquiry.ProductRevisionLevel, sizeof(inquiry.ProductRevisionLevel) );
   info.ProductRevisionLevel[sizeof(inquiry.ProductRevisionLevel)] = '\0';
   return true;
}

bool USB_DSC::CloseCamera() const
{
   return SendCmd(USB_CLOSE, NULL, 0);
}

struct TSendStartIO
{
   dword length;
   word code; // 1
   word code2; // 0x9FFF
   dword _unknown[2]; // for sizeof == 16
};

static TSendStartIO _Start_IO = { 16, 1, 0x9FFF };

bool USB_DSC::usbStartIO() const
{
   return SendCmd(USB_START_IO, &_Start_IO, sizeof(_Start_IO) );
}

bool USB_DSC::usbSendCmd(DSC_COMMAND_ENUM type, int reg) const
{
   struct TSendCmd
   {
      dword length;
      word  code; // 2
      word  code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      word check_sum;
   } send;

   memset(&send, 0, sizeof(send));

   send.length = sizeof(send);
   send.code = 2;

   send.magic = 0x431B;
   send.size = 2;
   send.command = dsc_command[type];
   send.reg = (byte)reg;
   send.check_sum = (word)((unsigned)send.command + (unsigned)send.reg);

   return SendCmd(USB_SEND_CMD, &send, send.length);
}

bool USB_DSC::usbSendCmdSetStr(int reg, const char* str) const
{
   struct TSendCmdSetStr
   {
      dword length;
      word  code; // 2
      word  code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      byte data[1];
   };

   int len = strlen(str);
   int size = sizeof(TSendCmdSetStr)-1+len+2;

   TSendCmdSetStr& send = *(TSendCmdSetStr*)new byte[size];
   memset(&send, 0, size);

   send.length = size;
   send.code = 2;

   send.magic = 0x431B;
   send.size = (word)(len+2);
   send.command = dsc_command[DSC_SET_VAR];
   send.reg = (byte)reg;

   unsigned sum = (unsigned)send.command + (unsigned)send.reg;
   for(int i=0; i<len; ++i)
   {
      send.data[i] = (byte)str[i];
      sum += (unsigned)send.data[i];
   }
   send.data[len] = (byte)(sum);
   send.data[len+1] = (byte)(sum>>8);

   bool ret = SendCmd(USB_SEND_CMD, &send, send.length);
   delete [] (byte*)&send;
   return ret;
}

bool USB_DSC::usbSendCmdWrite(int reg, const void* data, unsigned sz) const
{
   struct TSendCmdWrite
   {
      dword length;
      word  code; // 2
      word  code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      byte data[2]; // two bytes checksum
   };

   int size = sizeof(TSendCmdWrite)+sz;

   TSendCmdWrite& send = *(TSendCmdWrite*)new byte[size];
   memset(&send, 0, size);

   send.length = size;
   send.code = 2;

   send.magic = (word)0x431B;
   send.size = (word)(sz+2);
   send.command = dsc_command[DSC_SET_VAR];
   send.reg = (byte)reg;

   unsigned sum = (unsigned)send.command + (unsigned)send.reg;
   for(int i=0; i<sz; ++i)
   {
      send.data[i] = ((byte*)data)[i];
      sum += (unsigned)send.data[i];
   }
   send.data[sz] = (byte)(sum);
   send.data[sz+1] = (byte)(sum>>8);

   bool ret = SendCmd(USB_SEND_CMD, &send, send.length);
   delete [] (byte*)&send;
   return ret;
}

bool USB_DSC::usbSendCmdWrite2(int cmd, const void* data, unsigned sz) const
{
   struct TSendCmdWrite2
   {
      dword length;
      word  code; // 2
      word  code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word cmd;
      word size;
      byte data[2]; // two bytes checksum
   };

   int size = sizeof(TSendCmdWrite2)+sz;

   TSendCmdWrite2& send = *(TSendCmdWrite2*)new byte[size];
   memset(&send, 0, size);

   send.length = size;
   send.code = 2;

   send.cmd = (word)cmd;
   send.size = (word)(sz+2);

   unsigned sum = 0;
   for(int i=0; i<sz; ++i)
   {
      send.data[i] = ((byte*)data)[i];
      sum += (unsigned)send.data[i];
   }
   send.data[sz] = (byte)(sum);
   send.data[sz+1] = (byte)(sum>>8);

   bool ret = SendCmd(USB_SEND_CMD, &send, send.length);
   delete [] (byte*)&send;
   return ret;
}

bool USB_DSC::usbSendCmdSetReg(int reg, unsigned value) const
{
   struct TSendCmdVar
   {
      dword length;
      word code; // 2
      word code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      dword value;
      word check_sum;
   } send;

   memset(&send, 0, sizeof(send));

   send.length = sizeof(send);
   send.code = 2;

   send.magic = 0x431B;
   send.size = 6;
   send.command = dsc_command[DSC_SET_REG];
   send.reg = (byte)reg;
   send.value = value;
   send.check_sum = (word)
    ( (unsigned)(value&0xFF)
    + (unsigned)((value>>8)&0xFF)
    + (unsigned)((value>>16)&0xFF)
    + (unsigned)((value>>24)&0xFF)
    + (unsigned)send.reg);

   return SendCmd(USB_SEND_CMD, &send, send.length);
}

bool USB_DSC::usbSpecCmd(int reg, int value) const
{
   struct TSendCmd
   {
      dword length;
      word code; // 2
      word code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      byte value;
      word check_sum;
   } send;

   memset(&send, 0, sizeof(send));

   send.length = sizeof(send);
   send.code = 2;

   send.magic = 0x431B;
   send.size = 3;
   send.command = dsc_command[DSC_COMMAND];
   send.reg = (byte)reg;
   send.value = (byte)value;
   send.check_sum = (word)(
      (unsigned)send.command +
      (unsigned)send.reg +
      (unsigned)send.value);

   return SendCmd(USB_SEND_CMD, &send, send.length);
}

bool USB_DSC::usbSendAck() const
{
   TCmdAck send;
   memset(&send, 0, sizeof(send));

   send.length = sizeof(send);
   send.code = 2;
   send.ack = 6;

   return SendCmd(USB_SEND_CMD, &send, send.length);
}

bool USB_DSC::usbAcceptResponce(void* buf, int size) const
{
   return SendCmd(USB_RCV_RESPONCE, buf, size);
}

bool USB_DSC::usbGetAck() const
{
   struct TResponceAck
   {
      dword length;
      dword code;
      word _unknown[3]; // for sizeof == 14
   } responce;

   if( !SendCmd(USB_RCV_ACK, &responce, sizeof(responce)) ) return false;

   return (responce.code == 0x9FFF0003);
}

bool USB_DSC::usbGetSize(unsigned& size) const
{
   struct TResponceSize
   {
      dword length;
      dword code;
      dword _unknown;
      dword responce_size;
   } responce;

   if( SendCmd(USB_RCV_SIZE, &responce, sizeof(responce) ) )
   {
      size = responce.responce_size;
      return true;
   }
   return false;
}



#ifdef DEBUG
#define CHECK(x) { /*dprintf(" %s:%d %s\n", __FILE__, __LINE__, #x); */ if( !(x) ) { dprintf(" %s:%d %s - fault\n", __FILE__, __LINE__, #x); return false; }}
#else
#define CHECK(x) if( !(x) ) { return false; }
#endif

bool USB_DSC::GetVar(int reg, TResponce& responce) const
{
   responce.free();

   struct TResponceVar
   {
      dword length;
      word code; // 1,2,3
      word code2; // 0,0x9FFF
      dword _unknown[14]; // for sizeof == 64

      byte cmd; // 3,2
      byte frame; // 0,...
      word len; //
      byte data[PACKET_SIZE+8];
   } resp;

   CHECK( usbStartIO() );
   CHECK( usbSendCmd(DSC_GET_VAR, reg) );
   CHECK( usbGetAck() );

   for(;;)
   {
      unsigned size;
      CHECK( usbStartIO() );
      CHECK( usbGetSize(size) );
      CHECK( size != 0 );
      CHECK( size <= sizeof(resp) );
      CHECK( usbAcceptResponce(&resp, size));
      CHECK( usbGetAck() );

      CHECK( usbStartIO() );
      CHECK( usbSendAck() );
      CHECK( usbGetAck() );

      if( resp.cmd == 3 )
      {
         responce.append(resp.data, resp.len);
         break;
      }

      CHECK( resp.cmd == 2 );

      responce.append(resp.data, resp.len);
   }
   return true;
}

bool USB_DSC::SetStr(int reg, const char* str) const
{
   CHECK( usbStartIO() );
   CHECK( usbSendCmdSetStr(reg, str) );
   CHECK( usbGetAck() );

   unsigned size;
   TCmdAck responce;

   CHECK( usbStartIO() );
   CHECK( usbGetSize(size) );
   CHECK( size == sizeof(responce) );
   CHECK( usbAcceptResponce(&responce, size) );
   CHECK( usbGetAck() );

   CHECK( responce.ack == 6 );

   return true;
}

bool USB_DSC::SetVar(int reg, const void* data, unsigned size) const
{
   for(int i=0; size; ++i )
   {
      int cmd = (((i-1)&255)<<8) + 3;
      int sz = size;
      if( sz > PACKET_SIZE )
      {
         sz = PACKET_SIZE;
         --cmd;
      }

      CHECK( usbStartIO() );
      CHECK( i == 0 ? usbSendCmdWrite(reg, data, sz) : usbSendCmdWrite2(cmd, data, sz) );
      CHECK( usbGetAck() );

      unsigned size;
      TCmdAck responce;

      CHECK( usbStartIO() );
      CHECK( usbGetSize(size) );
      CHECK( size == sizeof(responce) );
      CHECK( usbAcceptResponce(&responce, size) );
      CHECK( usbGetAck() );

      CHECK( responce.ack == 6 );

      if( (size -= sz) == 0 ) break;
      data = (const void*)((unsigned)data + sz);
   }
   return true;
}

bool USB_DSC::GetReg(int reg, unsigned& value) const
{
   CHECK( usbStartIO() );
   CHECK( usbSendCmd(DSC_GET_REG, reg) );
   CHECK( usbGetAck() );

   value = -1;
   unsigned size;

   struct TResponceCmdGetReg
   {
      dword length;
      word code; // 1,2,3
      word code2; // 0,0x9FFF
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      dword value;
      word check_sum;
   } responce;

   CHECK( usbStartIO() );
   CHECK( usbGetSize(size) );
   CHECK( size <= sizeof(responce) );
   CHECK( usbAcceptResponce(&responce, size) );
   CHECK( usbGetAck() );

   if( size == sizeof(TCmdAck) )
      DOUT(("GetReg(%02X): responce=%02X\n", reg, ((TCmdAck*)&responce)->ack));

   CHECK( usbStartIO() );
   CHECK( usbSendAck() );
   CHECK( usbGetAck() );

   CHECK( size == sizeof(responce) );

   value = responce.value;

   return true;
}

bool USB_DSC::SpecCmd(int reg, unsigned value, unsigned timeout) const
{
   CHECK( usbStartIO() );
   CHECK( usbSpecCmd(reg, value) );
   CHECK( usbGetAck() );

   unsigned size;
   TCmdAck responce;

   CHECK( usbStartIO() );
   CHECK( usbGetSize(size) );
   CHECK( size == sizeof(responce) );
   CHECK( usbAcceptResponce(&responce, size) );
   CHECK( usbGetAck() );

   CHECK( responce.ack == 6 );

   if( timeout )
      Sleep( timeout );

   return true;
}

bool USB_DSC::SetReg(int reg, unsigned value) const
{
   CHECK( usbStartIO() );
   CHECK( usbSendCmdSetReg(reg, value) );
   CHECK( usbGetAck() );

   unsigned size;
   TCmdAck responce;

   CHECK( usbStartIO() );
   CHECK( usbGetSize(size) );
   CHECK( size == sizeof(responce) );
   CHECK( usbAcceptResponce(&responce, size) );
   CHECK( usbGetAck() );

   if( responce.ack == 6 ) return true;

   DOUT(("SetReg(%02X): responce=%02X\n", reg, responce.ack));
   return false;
}
////////////////////////////////////////////////////////////////////////////////

struct TScsiPassThrough
{
    word Length;
    byte ScsiStatus;
    byte PathId;
    byte TargetId;
    byte Lun;
    byte CdbLength;
    byte SenseInfoLength;
    byte DataIn;
    byte _align[3];
    dword DataTransferLength;
    dword TimeOutValue;
    void* DataBuffer;
    dword SenseInfoOffset;

    union
    {
       byte Cdb[16];

       // sizeof *this == 128 bytes
       byte buffer[128-7*4];

       TCdbCmd Cmd;
       TCdbCmdUSB USB;
    };
};

#define THROUGH_DIRECT

#ifdef THROUGH_DIRECT

static bool SendCmd_to_DSC_IoCtl_Pass_Through_Direct(HANDLE hDSC, TScsiCommand& command, void* buf, unsigned size)
{
   TScsiPassThrough data;
   memset(&data, 0, sizeof(data));

   data.Length = 44;
   //data.ScsiStatus = 0;
   //data.PathId = 0;
   data.TargetId = 1;
   //data.Lun = 0;
   data.CdbLength = 16;
   data.SenseInfoLength = 24;
   data.DataIn = command.data_in;
   data.DataTransferLength = size;
   data.TimeOutValue = timeout ? timeout : command.timeout;
   data.DataBuffer = buf;
   data.SenseInfoOffset = 48;

   if(data.DataIn)
      memset(buf, 0, size);

   if( command.command == 0x12 )
   {
      data.CdbLength = 12;
      data.Cmd.command = command.command;
      data.Cmd.size = size;
   }
   else
   {
      data.USB.command = command.command;
      data.USB.size_lo = (byte)(size);
      data.USB.size_hi = (byte)(size >> 8);
   }

   DWORD BytesReturned;
   return 0 != DeviceIoControl( hDSC, IOCTL_SCSI_PASS_THROUGH_DIRECT,
      &data, sizeof(data), &data, sizeof(data), &BytesReturned, NULL );
}

#else

static bool SendCmd_to_DSC_IoCtl_Pass_Through(HANDLE hDSC, TScsiCommand& command, void* buf, unsigned size)
{
   unsigned blk_size = 0x50+size;
   TScsiPassThrough* data = (TScsiPassThrough*)new char[blk_size];
   memset(data, 0, blk_size);

   data->Length = 44;
   //data->ScsiStatus = 0;
   //data->PathId = 0;
   data->TargetId = 1;
   //data->Lun = 0;
   data->CdbLength = 16;
   data->SenseInfoLength = 20;
   data->DataIn = command.data_in;
   data->DataTransferLength = size;
   data->TimeOutValue = timeout ? timeout : command.timeout;
   data->DataBuffer = (void*)0x50;
   data->SenseInfoOffset = 48;

   if(data->DataIn)
      memset((char*)data+0x50, 0, size);
   else
      memcpy((char*)data+0x50, buf, size);

   if( command.command == 0x12 )
   {
      data->CdbLength = 6;
      data->Cmd.command = command.command;
      data->Cmd.size = size;
   }
   else
   {
      data->USB.command = command.command;
      data->USB.size_lo = (byte)(size);
      data->USB.size_hi = (byte)(size >> 8);
   }

   DWORD BytesReturned;
   BOOL result = DeviceIoControl( hDSC, IOCTL_SCSI_PASS_THROUGH,
      data, (data->DataIn?44:blk_size), data, blk_size, &BytesReturned, NULL );

   if(data->DataIn)
      memcpy(buf, (char*)data+0x50, size);

   return 0 != result;
}

#endif//THROUGH_DIRECT

typedef bool ( * TSendCmd_to_DSC_IoCtl)(HANDLE, TScsiCommand&, void*, unsigned);

#ifdef THROUGH_DIRECT
static TSendCmd_to_DSC_IoCtl SendCmd_to_DSC_IoCtl = SendCmd_to_DSC_IoCtl_Pass_Through_Direct;
#else
static TSendCmd_to_DSC_IoCtl SendCmd_to_DSC_IoCtl = SendCmd_to_DSC_IoCtl_Pass_Through;
#endif


static bool GetDeviceName(HANDLE hDSC, TInquiryData& inquiry)
{
   return SendCmd_to_DSC_IoCtl(hDSC, inquiry_command, &inquiry, sizeof(inquiry));
}

class IOCTL_DSC : public USB_DSC
{
   HANDLE handle;

public:

   IOCTL_DSC(HANDLE _handle, char drive, VENDOR_ENUM vendor) :
      USB_DSC("USB\\A:", vendor),
      handle(_handle)
   {
      comment[4] = drive;
      DOUT(("Create IOCTL_DSC %s\n", comment))
   }

private:
   HANDLE GetHandle() const { return handle; }

   virtual ~IOCTL_DSC();
   virtual bool SendCmd(USB_COMMAND_ENUM i_command, void* buf, unsigned size) const;
};

bool IOCTL_DSC::SendCmd(USB_COMMAND_ENUM i_command, void* buf, unsigned size) const
{
   return SendCmd_to_DSC_IoCtl( GetHandle(), command_table[i_command], buf, size);
}

IOCTL_DSC::~IOCTL_DSC()
{
   DOUT(("Destroy IOCTL_DSC %s\n", comment))

   CloseCamera();

   ::CloseHandle( GetHandle() );
}

static char DSC_DRIVE[] = "\\\\.\\C:";
static char DSC_PATH[] = "C:\\";

DSC* OpenUSB(char camera_drive)
{
   DOUT(("OpenUSB IOCTL: %c\n", camera_drive))

   DSC_DRIVE[4] = camera_drive;

   HANDLE hDSC = CreateFile(DSC_DRIVE, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
      NULL, OPEN_EXISTING, 0, NULL );

   if( INVALID_HANDLE_VALUE == hDSC ) {
      DOUT(("CreateFile(%s): error=%u\n", DSC_DRIVE, GetLastError()))
      return NULL;
   }

   return new IOCTL_DSC( hDSC, DSC_DRIVE[4], default_vendor );
}

DSC* OpenUSB()
{
   DOUT(("OpenUSB IOCTL\n"))
   DWORD drives = GetLogicalDrives();

   for(int i=2; i < 26; ++i)
   {
      if( drives & (1<<i) )
      {
         DSC_DRIVE[4] =
         DSC_PATH[0] = (char)('A' + i);

         UINT drive_type = GetDriveType( DSC_PATH );
         DOUT(("GetDriveType(%s): %u\n", DSC_PATH, drive_type))

         if( drive_type != DRIVE_REMOVABLE ) {
            DOUT(("It is not removable drive\n"))
            continue;
         }

         HANDLE hDSC = CreateFile(DSC_DRIVE, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
            NULL, OPEN_EXISTING, 0, NULL );

         if( hDSC == INVALID_HANDLE_VALUE ) {
            DOUT(("CreateFile(%s): error=%u\n", DSC_DRIVE, GetLastError()))
            continue;
         }

         TInquiryData inquiry;
         bool ret = GetDeviceName(hDSC, inquiry);

         DOUT(("IOCTL %s [%8.8s][%16.16s][%4.4s]\n", DSC_DRIVE, inquiry.VendorId, inquiry.ProductId, inquiry.ProductRevisionLevel))

         if( ret )
         {
            if( IsNikonDSC( inquiry ) )
            {
               DOUT(("It is a Nikon Digital Camera\n"))
               return new IOCTL_DSC( hDSC, DSC_DRIVE[4], NIKON );
            }

            if( IsOlympusDSC( inquiry ) )
            {
               DOUT(("It is a Olympus Digital Camera\n"))
               return new IOCTL_DSC( hDSC, DSC_DRIVE[4], OLYMPUS );
            }
         }

         DOUT(("It is unknown device\n"))
         CloseHandle( hDSC );
      }
   }
   DOUT(("OpenUSB IOCTL: camera is not found\n"))
   return NULL;
}

bool CheckUSB(char camera_drive, char* buf, unsigned size)
{
   DOUT(("CheckUSB %c:\n", camera_drive))

   int len = wsprintf( buf, "Open drive %c: ", camera_drive );
   buf += len;
   size -= len;

   DSC_DRIVE[4] = camera_drive;

   SetLastError( NOERROR );

   HANDLE hDSC = CreateFile(DSC_DRIVE, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
      NULL, OPEN_EXISTING, 0, NULL );

   if( INVALID_HANDLE_VALUE == hDSC )
   {
      DOUT(("IOCTL %s error=%u\n", DSC_DRIVE, GetLastError()))
      
      len = wsprintf( buf, "\r\nERROR: " );
      buf += len;
      size -= len;

      len = FormatMessage( FORMAT_MESSAGE_MAX_WIDTH_MASK	|
         FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
         NULL, GetLastError(), 0, buf, size-2, NULL );

      buf += len;

      strcpy( buf, "\r\n" );
      return false;
   }
   else
   {
      len = wsprintf( buf, "SUCCESS \r\n" );
      buf += len;
      size -= len;
   }

   len = wsprintf( buf, "Inquiry drive:\r\n" );
   buf += len;
   size -= len;

   TInquiryData inquiry;

   SetLastError( NOERROR );

   bool ret = GetDeviceName(hDSC, inquiry);

   if( ret )
   {
      DOUT(("IOCTL %s [%8.8s][%16.16s][%4.4s]\n", DSC_DRIVE, inquiry.VendorId, inquiry.ProductId, inquiry.ProductRevisionLevel))

      wsprintf( buf, "%sDevice: %.8s  %.16s %.4s\r\n",
         inquiry.RemovableMedia ? "Removable ":"",
         inquiry.VendorId, inquiry.ProductId, inquiry.ProductRevisionLevel);
   }
   else
   {
      DOUT(("IOCTL %s error=%u\n", DSC_DRIVE, GetLastError()))

      len = wsprintf( buf, "ERROR: " );
      buf += len;
      size -= len;

      len = FormatMessage( FORMAT_MESSAGE_MAX_WIDTH_MASK	|
         FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
         NULL, GetLastError(), 0, buf, size-2, NULL );

      buf += len;

      strcpy( buf, "\r\n" );
   }

   CloseHandle( hDSC );
   return true;
}

const char* IOCTL_DriveName::GetNext(unsigned& param)
{
   if( i == -1 )
   {
      i = 0;
      drives = GetLogicalDrives();
   }

   for(i; i < 26; ++i)
   {
      if( drives & (1<<i) )
      {
         char c = (char)('A' + i);
         wsprintf(drive, "%c:", c);
         param = c | PARAM_USB_IOCTL;
         ++i;
         return drive;
      }
   }
   return NULL;
}

////////////////////////////////////////////////////////////////////////////////

#ifdef SUPPORT_WINDOWS_98SE

#include <wnaspi32.h>

#ifndef SCSIOP_INQUIRY
#define SCSIOP_INQUIRY 0x0012
#endif

typedef DWORD (__cdecl* TGetASPI32SupportInfo)(void);
typedef DWORD (__cdecl* TSendASPI32Command)(LPSRB);

int IsASPI;
HINSTANCE hASPI;
TGetASPI32SupportInfo pGetASPI32SupportInfo;
TSendASPI32Command pSendASPI32Command;

void wait() {}

bool GetDeviceName(int host, int scsi_id, int lun, TInquiryData& inquiry)
{
   memset( &inquiry, 0, sizeof(TInquiryData) );

   SRB_ExecSCSICmd cmd;

   memset( &cmd, 0, sizeof cmd );

   cmd.SRB_Cmd = SC_EXEC_SCSI_CMD;
   cmd.SRB_HaId = (BYTE) host;
   cmd.SRB_Flags = SRB_DIR_IN;
   cmd.SRB_Target = (BYTE) scsi_id;
   cmd.SRB_Lun = (BYTE) lun;
   cmd.SRB_BufLen = sizeof(TInquiryData);
   cmd.SRB_BufPointer = (BYTE*)&inquiry;
   cmd.SRB_SenseLen = SENSE_LEN;
   cmd.SRB_CDBLen = 6;
   cmd.CDBByte[0] = SCSIOP_INQUIRY;
   *(WORD*)(&cmd.CDBByte[4]) = (WORD)sizeof(TInquiryData);

   pSendASPI32Command(&cmd);

   while( cmd.SRB_Status == SS_PENDING )
      wait();

   return cmd.SRB_Status == SS_COMP;
}

bool SendCmd_to_DSC_SCSI(int host, int id, int lun, TScsiCommand& command, void* buf, unsigned size)
{
   SRB_ExecSCSICmd cmd;
   memset(&cmd, 0, sizeof(cmd));

   cmd.SRB_Cmd = SC_EXEC_SCSI_CMD;
   cmd.SRB_HaId = (BYTE) host;
   cmd.SRB_Flags = 0;
   cmd.SRB_Target = (BYTE) id;
   cmd.SRB_Lun = (BYTE) lun;
   cmd.SRB_BufLen = size;
   cmd.SRB_BufPointer = (BYTE*)buf;
   cmd.SRB_SenseLen = SENSE_LEN;
   cmd.SRB_CDBLen = 16;

   cmd.SRB_Flags |= (BYTE)(command.data_in ? SRB_DIR_IN : SRB_DIR_OUT);

   if(command.data_in)
      memset(buf, 0, size);

   if( command.command == SCSIOP_INQUIRY )
   {
      cmd.SRB_CDBLen = 12;
      TCdbCmd& Cmd = *(TCdbCmd*)(&cmd.CDBByte[0]);

      Cmd.command = command.command;
      Cmd.size = size;
   }
   else
   {
      TCdbCmdUSB& Cmd = *(TCdbCmdUSB*)(&cmd.CDBByte[0]);
      Cmd.command = command.command;
      Cmd.size_lo = (byte)(size);
      Cmd.size_hi = (byte)(size >> 8);
   }

   pSendASPI32Command(&cmd);

   while( cmd.SRB_Status == SS_PENDING )
      wait();

   return cmd.SRB_Status == SS_COMP;
}

class ASPI_DSC : public USB_DSC
{
   int host, id, lun;

public:
   ASPI_DSC(int _host, int _id, int _lun, VENDOR_ENUM vendor)
   : USB_DSC("USB:", vendor), host(_host), id(_id), lun(_lun)
   {
      DOUT(("Create ASPI_DSC %s\n", comment))
   }

private:
   virtual ~ASPI_DSC();
   virtual bool SendCmd(USB_COMMAND_ENUM i_command, void* buf, unsigned size) const;
};

ASPI_DSC::~ASPI_DSC()
{
   DOUT(("Destroy ASPI_DSC %s\n", comment))
   CloseCamera();
}

bool ASPI_DSC::SendCmd(USB_COMMAND_ENUM i_command, void* buf, unsigned size) const
{
   return SendCmd_to_DSC_SCSI( host, id, lun, command_table[i_command], buf, size);
}

bool InitASPI()
{
   if( IsASPI != 0 )
      return IsASPI == 1;

   IsASPI = -1;
   hASPI = LoadLibrary("WNASPI32.DLL");
   
   if( !hASPI ) {
      DOUT(("LoadLibrary(WNASPI32.DLL): error=%u\n", GetLastError()))
      return false;
   }
   
   pGetASPI32SupportInfo =
      (TGetASPI32SupportInfo)GetProcAddress(hASPI, "GetASPI32SupportInfo");

   if( !pGetASPI32SupportInfo ) {
      DOUT(("GetASPI32SupportInfo: error=%u\n", GetLastError()))
      return false;
   }

   pSendASPI32Command =
      (TSendASPI32Command)GetProcAddress(hASPI, "SendASPI32Command");

   if( !pSendASPI32Command ) {
      DOUT(("SendASPI32Command: error=%u\n", GetLastError()))
      return false;
   }
   IsASPI = 1;
   return true;
}

DSC* OpenUSB_ASPI()
{
   DOUT(("OpenUSB ASPI\n"))

   if( !InitASPI() ) return NULL;

   DWORD support = pGetASPI32SupportInfo();

   int status = (support>>8) & 255;
   if( status != SS_COMP ) {
      DOUT(("GetASPI32SupportInfo: status=%08X\n", status))
      return NULL;
   }

   int nHosts = support & 255;

   for(int host = 0; host < nHosts; ++host)
   {
      SRB_HAInquiry hai;
      memset(&hai, 0, sizeof hai );

      hai.SRB_Cmd = SC_HA_INQUIRY;
      hai.SRB_HaId = (BYTE)host;          // ASPI host adapter number

      DWORD ret;
      ret = pSendASPI32Command(&hai);

      if( ret != SS_COMP ) continue;
      if( hai.SRB_Status != SS_COMP ) continue;

      DOUT(("SCSI Host #%d: %s\n", host, hai.HA_Identifier))

      for(int scsi_id = 0; scsi_id<8; ++scsi_id)
      for(int lun = 0; lun<8; ++lun)
      {
         SRB_GDEVBlock device;

         memset(&device, 0, sizeof device );
         device.SRB_Cmd = SC_GET_DEV_TYPE;
         device.SRB_HaId = (BYTE)host;
         device.SRB_Target = (BYTE)scsi_id;
         device.SRB_Lun = (BYTE)lun;

         ret = pSendASPI32Command(&device);

         if( ret != SS_COMP ) continue;
         if( device.SRB_Status == SS_COMP )
         {
            TInquiryData inquiry;
            bool ret = GetDeviceName(host, scsi_id, lun, inquiry);

            DOUT(("SCSI id:%d, lun:%d [%8.8s][%16.16s][%4.4s]\n", scsi_id, lun, inquiry.VendorId, inquiry.ProductId, inquiry.ProductRevisionLevel))

            if( ret )
            {
               if( IsNikonDSC( inquiry ) )
               {
                  DOUT(("It is a Nikon Digital Camera\n"))
                  return new ASPI_DSC( host, scsi_id, lun, NIKON );
               }

               if( IsOlympusDSC( inquiry ) )
               {
                  DOUT(("It is a Olympus Digital Camera\n"))
                  return new ASPI_DSC( host, scsi_id, lun, OLYMPUS );
               }
            }
         }
      }
   }
   DOUT(("OpenUSB ASPI: camera is not found\n"))
   return NULL;
}

DSC* OpenUSB_ASPI(int host, int scsi_id, int lun)
{
   DOUT(("OpenUSB ASPI: host:%d id:%d lun:%d\n", host, scsi_id, lun))
   TInquiryData inquiry;
   bool ret = GetDeviceName(host, scsi_id, lun, inquiry);

   if( !ret ) {
      DOUT(("OpenUSB ASPI: camera is not found\n"))
      return NULL;
   }                                         
   return new ASPI_DSC( host, scsi_id, lun, default_vendor );
}

bool CheckUSB_ASPI(int host, int scsi_id, int lun, char* buf, unsigned size)
{
   DOUT(("CheckUSB_ASPI ha:%d id:%d, lun:%d\n", host, scsi_id, lun))

   int len = wsprintf( buf, "Open device host:%d id:%d lun:%d\r\n", host, scsi_id, lun );
   buf += len;
   size -= len;

   SetLastError( NOERROR );

   if( !InitASPI() ) {
      wsprintf( buf, "ASPI32 is not installed\r\n" );
      return NULL;
   }

   DWORD support = pGetASPI32SupportInfo();

   int status = (support>>8) & 255;
   if( status != SS_COMP ) {
      DOUT(("GetASPI32SupportInfo: status=%08X\n", status))
      wsprintf( buf, "Error Status: 0x%02X\r\n", status );
      return NULL;
   }

   SRB_GDEVBlock device;

   memset(&device, 0, sizeof device );
   device.SRB_Cmd = SC_GET_DEV_TYPE;
   device.SRB_HaId = (BYTE)host;
   device.SRB_Target = (BYTE)scsi_id;
   device.SRB_Lun = (BYTE)lun;

   pSendASPI32Command(&device);

   if( device.SRB_Status != SS_COMP )
   {
      DOUT(("SC_GET_DEV_TYPE ha:%d id:%d, lun:%d status=0x%02X\n", host, scsi_id, lun, device.SRB_Status))
      wsprintf( buf, "Error Status: 0x%02X\r\n", device.SRB_Status );
      return true;
   }
   len = wsprintf( buf, "Inquiry drive:\r\n" );
   buf += len;
   size -= len;

   SetLastError( NOERROR );

   TInquiryData inquiry;

   memset( &inquiry, 0, sizeof(TInquiryData) );

   SRB_ExecSCSICmd cmd;

   memset( &cmd, 0, sizeof cmd );

   cmd.SRB_Cmd = SC_EXEC_SCSI_CMD;
   cmd.SRB_HaId = (BYTE) host;
   cmd.SRB_Flags = SRB_DIR_IN;
   cmd.SRB_Target = (BYTE) scsi_id;
   cmd.SRB_Lun = (BYTE) lun;
   cmd.SRB_BufLen = sizeof(TInquiryData);
   cmd.SRB_BufPointer = (BYTE*)&inquiry;
   cmd.SRB_SenseLen = SENSE_LEN;
   cmd.SRB_CDBLen = 6;
   cmd.CDBByte[0] = SCSIOP_INQUIRY;
   *(WORD*)(&cmd.CDBByte[4]) = (WORD)sizeof(TInquiryData);

   pSendASPI32Command(&cmd);

   while( cmd.SRB_Status == SS_PENDING )
      wait();

   if( cmd.SRB_Status != SS_COMP )
   {
      DOUT(("SCSIOP_INQUIRY ha:%d id:%d, lun:%d status=0x%02X\n", host, scsi_id, lun, cmd.SRB_Status))
      wsprintf( buf, "Error Status: 0x%02X\r\n", cmd.SRB_Status );
      return true;
   }

   DOUT(("SCSIOP_INQUIRY ha:%d id:%d, lun:%d [%8.8s][%16.16s][%4.4s]\n", host, scsi_id, lun, inquiry.VendorId, inquiry.ProductId, inquiry.ProductRevisionLevel))

   wsprintf( buf, "%sDevice: %.8s  %.16s %.4s\r\n",
      inquiry.RemovableMedia ? "Removable ":"",
      inquiry.VendorId, inquiry.ProductId, inquiry.ProductRevisionLevel);

   size;
   return true;
}

const char* ASPI_HostName::GetNext(unsigned& param)
{
   if( HaId == -1 )
   {
      if( !InitASPI() )
         return NULL;

      DWORD support = pGetASPI32SupportInfo();

      int status = (support>>8) & 255;
      if( status != SS_COMP )
         return NULL;

      HaId = SCSI_Id = Lun = 0;
      nHosts = support & 255;
   }

   bool ok = false;
   for(HaId; HaId < nHosts; ++HaId)
   {
      SRB_HAInquiry hai;
      memset(&hai, 0, sizeof hai );

      hai.SRB_Cmd = SC_HA_INQUIRY;
      hai.SRB_HaId = (BYTE)HaId;          // ASPI host adapter number

      DWORD ret;
      ret = pSendASPI32Command(&hai);

      if( ret != SS_COMP ) continue;
      if( hai.SRB_Status != SS_COMP ) continue;

      for(SCSI_Id; SCSI_Id<8; ++SCSI_Id)
      {
         for(Lun; Lun<8; ++Lun)
         {
            if( ok )
               return host;

            SRB_GDEVBlock device;

            memset(&device, 0, sizeof device );
            device.SRB_Cmd = SC_GET_DEV_TYPE;
            device.SRB_HaId = (BYTE)HaId;
            device.SRB_Target = (BYTE)SCSI_Id;
            device.SRB_Lun = (BYTE)Lun;

            ret = pSendASPI32Command(&device);

            if( ret != SS_COMP ) continue;
            if( device.SRB_Status == SS_COMP )
            {
               wsprintf(host, "HaId:%u Id:%u Lun:%u", HaId, SCSI_Id, Lun);
               param = (HaId << 16) | (SCSI_Id << 8) | Lun | PARAM_USB_ASPI;
               ok = true;
            }
         }
         Lun = 0;
      }
      SCSI_Id = 0;
   }

   if( ok )
      return host;

   return NULL;
}
#endif

